#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import uuid

import numpy as np
from caffe2.python import workspace
from caffe2.python.predictor import predictor_exporter

from .builtin_task import register_builtin_tasks
from .config.component import create_featurizer
from .data.featurizer import InputRecord
from .utils.onnx_utils import CAFFE2_DB_TYPE, convert_caffe2_blob_name


register_builtin_tasks()


def _predict(workspace_id, feature_config, predict_net, featurizer, input):
    workspace.SwitchWorkspace(workspace_id)
    features = featurizer.featurize(InputRecord(**input))
    if feature_config.word_feat:
        for blob_name in feature_config.word_feat.export_input_names:
            converted_blob_name = convert_caffe2_blob_name(blob_name)
            workspace.blobs[converted_blob_name] = np.array(
                [features.tokens], dtype=str
            )
        workspace.blobs["tokens_lens"] = np.array([len(features.tokens)], dtype=np.int_)
    if feature_config.dict_feat:
        dict_feats, weights, lens = feature_config.dict_feat.export_input_names
        converted_dict_blob_name = convert_caffe2_blob_name(dict_feats)
        workspace.blobs[converted_dict_blob_name] = np.array(
            [features.gazetteer_feats], dtype=str
        )
        workspace.blobs[weights] = np.array(
            [features.gazetteer_feat_weights], dtype=np.float32
        )
        workspace.blobs[lens] = np.array(features.gazetteer_feat_lengths, dtype=np.int_)

    if feature_config.char_feat:
        for blob_name in feature_config.char_feat.export_input_names:
            converted_blob_name = convert_caffe2_blob_name(blob_name)
            workspace.blobs[converted_blob_name] = np.array(
                [features.characters], dtype=str
            )

    workspace.RunNet(predict_net)
    return {
        str(blob): workspace.blobs[blob][0] for blob in predict_net.external_outputs
    }


def create_predictor(config, model_file=None):
    workspace_id = str(uuid.uuid4())
    workspace.SwitchWorkspace(workspace_id, True)
    predict_net = predictor_exporter.prepare_prediction_net(
        filename=model_file or config.export_caffe2_path, db_type=CAFFE2_DB_TYPE
    )

    task = config.task
    feature_config = task.features
    featurizer = create_featurizer(task.featurizer, feature_config)

    return lambda input: _predict(
        workspace_id, feature_config, predict_net, featurizer, input
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import os
import signal

import torch
import torch.distributed as dist_c10d


def dist_init(
    distributed_rank: int, world_size: int, init_method: str, backend: str = "nccl"
):
    if init_method and world_size > 1 and torch.cuda.is_available():
        dist_c10d.init_process_group(
            backend=backend,
            init_method=init_method,
            world_size=world_size,
            rank=distributed_rank,
        )

        if distributed_rank != 0:
            suppress_output()


def suppress_output():
    import builtins as __builtin__

    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        # force print the result when kwargs contains force and value is True
        if kwargs.pop("force", False):
            builtin_print(*args, **kwargs)

    __builtin__.print = print



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from enum import Enum
from typing import List, Optional, Tuple

import numpy as np
import pytext.utils.cuda_utils as cuda_utils
import torch
import torch.nn as nn
import torch.nn.functional as F
from pytext.config import ConfigBase
from pytext.config.component import Component, ComponentType
from pytext.data import CommonMetadata
from pytext.models import Model
from pytext.models.embeddings import EmbeddingList
from pytext.models.representations.bilstm import BiLSTM
from pytext.models.semantic_parsers.rnng.rnng_data_structures import (
    CompositionalNN,
    CompositionalSummationNN,
    CompositionFunction,
    Element,
    ParserState,
)


class CompositionalType(Enum):
    """Whether to use summation of the vectors or a BiLSTM based composition to
     generate embedding for a subtree"""

    BLSTM = "blstm"
    SUM = "sum"


class AblationParams(ConfigBase):
    """Different ablation parameters.
    use_last_open_NT_feature uses the last open non-terminal as a 1-hot feature
    when computing representation for the action classifier

    Attributes:
    use_buffer: whether to use the buffer LSTM
    use_stack: whether to use the stack LSTM
    use_action: whether to use the action LSTM
    """

    use_buffer: bool = True
    use_stack: bool = True
    use_action: bool = True
    use_last_open_NT_feature: bool = False


class RNNGConstraints(ConfigBase):
    """Constraints when computing valid actions.

    intent_slot_nesting: for the intent slot models, the top level non-terminal
    has to be an intent, an intent can only have slot non-terminals as
    children and vice-versa.

    ignore_loss_for_unsupported and no_slots_inside_unsupported:
    if the data has "unsupported" label, that is if the label has a substring
    "unsupported" in it, these flags control whether or not to compute loss or
    predict slots in side the unsupported intent
    """

    intent_slot_nesting: bool = True
    ignore_loss_for_unsupported: bool = False
    no_slots_inside_unsupported: bool = True


class RNNGParser(Model, Component):
    """
    The Recurrent Neural Network Grammar (RNNG) parser from
    Dyer et al.: https://arxiv.org/abs/1602.07776 and
    Gupta et al.: https://arxiv.org/abs/1810.07942d.
    RNNG is a neural constituency parsing algorithm that
    explicitly models compositional structure of a sentence.
    It is able to learn about hierarchical relationship among the words and
    phrases in a given sentence thereby learning the underlying tree structure.
    The paper proposes generative as well as discriminative approaches.
    In PyText we have implemented the discriminative approach for modeling
    intent slot models.
    It is a top-down shift-reduce parser than can output
    trees with non-terminals (intent and slot labels) and terminals (tokens)
    """

    __COMPONENT_TYPE__ = ComponentType.MODEL

    class Config(ConfigBase):
        version: int = 0
        lstm: BiLSTM.Config = BiLSTM.Config()
        ablation: AblationParams = AblationParams()
        constraints: RNNGConstraints = RNNGConstraints()
        max_open_NT: int = 10
        dropout: float = 0.1
        compositional_type: CompositionalType = CompositionalType.BLSTM

    @classmethod
    def from_config(cls, model_config, feature_config, metadata: CommonMetadata):
        if model_config.compositional_type == CompositionalType.SUM:
            p_compositional = CompositionalSummationNN(
                lstm_dim=model_config.lstm.lstm_dim
            )
        elif model_config.compositional_type == CompositionalType.BLSTM:
            p_compositional = CompositionalNN(lstm_dim=model_config.lstm.lstm_dim)
        else:
            raise ValueError(
                "Cannot understand compositional flag {}".format(
                    model_config.compositional_type
                )
            )

        return cls(
            ablation=model_config.ablation,
            constraints=model_config.constraints,
            lstm_num_layers=model_config.lstm.num_layers,
            lstm_dim=model_config.lstm.lstm_dim,
            max_open_NT=model_config.max_open_NT,
            dropout=model_config.dropout,
            actions_vocab=metadata.actions_vocab,
            shift_idx=metadata.shift_idx,
            reduce_idx=metadata.reduce_idx,
            ignore_subNTs_roots=metadata.ignore_subNTs_roots,
            valid_NT_idxs=metadata.valid_NT_idxs,
            valid_IN_idxs=metadata.valid_IN_idxs,
            valid_SL_idxs=metadata.valid_SL_idxs,
            embedding=Model.create_embedding(feature_config, metadata=metadata),
            p_compositional=p_compositional,
        )

    def __init__(
        self,
        ablation: AblationParams,
        constraints: RNNGConstraints,
        lstm_num_layers: int,
        lstm_dim: int,
        max_open_NT: int,
        dropout: float,
        actions_vocab,
        shift_idx: int,
        reduce_idx: int,
        ignore_subNTs_roots: List[int],
        valid_NT_idxs: List[int],
        valid_IN_idxs: List[int],
        valid_SL_idxs: List[int],
        embedding: EmbeddingList,
        p_compositional: CompositionFunction,
    ) -> None:
        """
        Initialize the model

        Args:
        ablation : AblationParams
            Features/RNNs to use
        constraints : RNNGConstraints
            Constraints to use when computing valid actions
        lstm_num_layers : int
            number of layers in the LSTMs
        lstm_dim : int
            size of LSTM
        max_open_NT : int
            number of maximum open non-terminals allowed on the stack.
            After that, the only valid actions are SHIFT and REDUCE
        dropout : float
            dropout parameter
        actions_vocab : Vocab (right now torchtext.vocab.Vocab)
            dictionary of actions
        shift_idx : int
            index of shift action
        reduce_idx : int
            index of reduce action
        ignore_subNTs_roots : List[int]
            for these top non-terminals, ignore loss for all subsequent actions
        valid_NT_idxs : List[int]
            indices of all non-terminals
        valid_IN_idxs : List[int]
            indices of intent non-terminals
        valid_SL_idxs : List[int]
            indices of slot non-terminals
        embedding : EmbeddingList
            embeddings for the tokens
        p_compositional : CompositionFunction
            Composition function to use to get embedding of a sub-tree


        Returns:
        None


        """

        nn.Module.__init__(self)

        self.embedding = embedding
        # self.embedding.config: FeatureConfig object cannot be pickled but,
        # we require the model to be pickled for passing from one worker process
        # for Hogwild training. Hence, setting the config to None
        self.embedding.config = None

        self.p_compositional = p_compositional
        self.ablation_use_last_open_NT_feature = ablation.use_last_open_NT_feature
        self.ablation_use_buffer = ablation.use_buffer
        self.ablation_use_stack = ablation.use_stack
        self.ablation_use_action = ablation.use_action

        self.constraints_intent_slot_nesting = constraints.intent_slot_nesting
        self.constraints_no_slots_inside_unsupported = (
            constraints.no_slots_inside_unsupported
        )
        self.constraints_ignore_loss_for_unsupported = (
            constraints.ignore_loss_for_unsupported
        )

        self.lstm_num_layers = lstm_num_layers
        self.lstm_dim = lstm_dim
        self.max_open_NT = max_open_NT
        self.actions_vocab = actions_vocab
        self.shift_idx = shift_idx
        self.reduce_idx = reduce_idx
        self.ignore_subNTs_roots = ignore_subNTs_roots
        self.valid_NT_idxs = valid_NT_idxs
        self.valid_IN_idxs = valid_IN_idxs
        self.valid_SL_idxs = valid_SL_idxs

        num_actions = len(actions_vocab)
        lstm_count = ablation.use_buffer + ablation.use_stack + ablation.use_action
        if lstm_count == 0:
            raise ValueError("Need atleast one of the LSTMs to be true")

        self.action_linear = nn.Sequential(
            nn.Linear(
                lstm_count * lstm_dim + num_actions * ablation.use_last_open_NT_feature,
                lstm_dim,
            ),
            nn.ReLU(),
            nn.Linear(lstm_dim, num_actions),
        )
        self.dropout_layer = nn.Dropout(p=dropout)
        self.buff_rnn = nn.LSTM(
            embedding.embedding_dim,
            lstm_dim,
            num_layers=lstm_num_layers,
            dropout=dropout,
        )
        self.stack_rnn = nn.LSTM(
            lstm_dim, lstm_dim, num_layers=lstm_num_layers, dropout=dropout
        )
        self.action_rnn = nn.LSTM(
            lstm_dim, lstm_dim, num_layers=lstm_num_layers, dropout=dropout
        )

        self.pempty_buffer_emb = nn.Parameter(torch.randn(1, lstm_dim))
        self.empty_stack_emb = nn.Parameter(torch.randn(1, lstm_dim))
        self.empty_action_emb = nn.Parameter(torch.randn(1, lstm_dim))

        self.actions_lookup = nn.Embedding(num_actions, lstm_dim)
        self.loss_func = nn.CrossEntropyLoss()

    def forward(
        self,
        tokens: torch.Tensor,
        seq_lens: torch.Tensor,
        dict_feat: Optional[Tuple[torch.Tensor, ...]] = None,
        actions: Optional[List[List[int]]] = None,
        beam_size: int = 1,
        topk: int = 1,
    ):
        """RNNG forward function.

        Args:
        tokens : torch.Tensor
            list of tokens
        seq_lens : torch.Tensor
            list of sequence lengths
        dict_feat : Optional[Tuple[torch.Tensor, ...]]
            dictionary or gazetter features for each token
        actions : Optional[List[List[int]]]
            Used only during training. Oracle actions for the instances.
        beam_size : int
            Beam size; used only during inference
        topk : int
            Number of top results from the method. if beam_size is 1 this is 1.


        Returns:
        if topk == 1
            tuple of list of predicted actions and list of corresponding scores
        else
            returns
            list of tuple of list of predicted actions and list of corresponding scores


        """
        if self.training:
            assert beam_size == 1, "beam_size must be 1 during training"
            assert actions is not None, "actions must be provided for training"
            actions_idx_rev = list(reversed(actions[0]))
        else:
            torch.manual_seed(0)

        beam_size = max(beam_size, 1)

        # Reverse the order of indices along last axis before embedding lookup.
        tokens_list_rev = torch.flip(tokens, [len(tokens.size()) - 1])
        dict_feat_rev = None
        if dict_feat:
            dict_ids, dict_weights, dict_lengths = dict_feat
            dict_ids_rev = torch.flip(dict_ids, [len(dict_ids.size()) - 1])
            dict_weights_rev = torch.flip(dict_weights, [len(dict_weights.size()) - 1])
            dict_lengths_rev = torch.flip(dict_lengths, [len(dict_lengths.size()) - 1])
            dict_feat_rev = (dict_ids_rev, dict_weights_rev, dict_lengths_rev)

        embedding_input = (
            [tokens_list_rev, dict_feat_rev]
            if dict_feat_rev is not None
            else [tokens_list_rev]
        )
        tok_embeddings = self.embedding(*embedding_input)

        # Batch size is always = 1. So we squeeze the batch_size dimension.
        tok_embeddings = tok_embeddings.squeeze(0)
        tokens_list_rev = tokens_list_rev.squeeze(0)

        initial_state = ParserState(self)
        for i in range(tok_embeddings.size()[0]):
            tok_embedding = tok_embeddings[i].unsqueeze(0)
            tok = tokens_list_rev[i]
            initial_state.buffer_stackrnn.push(tok_embedding, Element(tok))

        beam = [initial_state]
        while beam and any(not state.finished() for state in beam):
            # Stores plans for expansion as (score, state, action)
            plans: List[Tuple[float, ParserState, int]] = []
            # Expand current beam states
            for state in beam:
                # Keep terminal states
                if state.finished():
                    plans.append((state.neg_prob, state, -1))
                    continue

                #  translating Expression p_t = affine_transform({pbias, S,
                #  stack_summary, B, buffer_summary, A, action_summary});
                stack = state.stack_stackrnn
                stack_summary = stack.embedding()
                action_summary = state.action_stackrnn.embedding()
                buffer_summary = state.buffer_stackrnn.embedding()
                if self.dropout_layer.p > 0:
                    stack_summary = self.dropout_layer(stack_summary)
                    action_summary = self.dropout_layer(action_summary)
                    buffer_summary = self.dropout_layer(buffer_summary)

                # feature for index of last open non-terminal
                last_open_NT_feature = torch.zeros(len(self.actions_vocab))
                open_NT_exists = state.num_open_NT > 0

                if (
                    len(stack) > 0
                    and open_NT_exists
                    and self.ablation_use_last_open_NT_feature
                ):
                    last_open_NT = None
                    try:
                        open_NT = state.is_open_NT[::-1].index(True)
                        last_open_NT = stack.ele_from_top(open_NT)
                    except ValueError:
                        pass
                    if last_open_NT:
                        last_open_NT_feature[last_open_NT.node] = 1.0
                last_open_NT_feature = last_open_NT_feature.unsqueeze(0)

                summaries = []
                if self.ablation_use_buffer:
                    summaries.append(buffer_summary)
                if self.ablation_use_stack:
                    summaries.append(stack_summary)
                if self.ablation_use_action:
                    summaries.append(action_summary)
                if self.ablation_use_last_open_NT_feature:
                    summaries.append(last_open_NT_feature)

                action_p = self.action_linear(torch.cat(summaries, dim=1))

                log_probs = F.log_softmax(action_p, dim=1)[0]

                for action in self.valid_actions(state):
                    plans.append((state.neg_prob - log_probs[action], state, action))

            beam = []
            # Take actions to regenerate the beam
            for neg_prob, state, predicted_action_idx in sorted(plans)[:beam_size]:
                # Skip terminal states
                if state.finished():
                    beam.append(state)
                    continue

                # Only branch out states when needed
                if beam_size > 1:
                    state = state.copy()

                state.predicted_actions_idx.append(predicted_action_idx)

                target_action_idx = predicted_action_idx
                if self.training:
                    assert (
                        len(actions_idx_rev) > 0
                    ), "Actions and tokens may not be in sync."
                    target_action_idx = actions_idx_rev[-1]
                    actions_idx_rev = actions_idx_rev[:-1]

                if (
                    self.constraints_ignore_loss_for_unsupported
                    and state.found_unsupported
                ):
                    pass
                else:
                    state.action_scores.append(action_p)

                action_embedding = self.actions_lookup(
                    cuda_utils.Variable(torch.LongTensor([target_action_idx]))
                )
                state.action_stackrnn.push(action_embedding, Element(target_action_idx))

                if target_action_idx == self.shift_idx:
                    state.is_open_NT.append(False)
                    tok_embedding, token = state.buffer_stackrnn.pop()
                    state.stack_stackrnn.push(tok_embedding, Element(token))
                elif target_action_idx == self.reduce_idx:
                    state.num_open_NT -= 1
                    popped_rep = []
                    nt_tree = []

                    while not state.is_open_NT[-1]:
                        assert len(state.stack_stackrnn) > 0, "How come stack is empty!"
                        state.is_open_NT.pop()
                        top_of_stack = state.stack_stackrnn.pop()
                        popped_rep.append(top_of_stack[0])
                        nt_tree.append(top_of_stack[1])

                    # pop the open NT and close it
                    top_of_stack = state.stack_stackrnn.pop()
                    popped_rep.append(top_of_stack[0])
                    nt_tree.append(top_of_stack[1])

                    state.is_open_NT.pop()
                    state.is_open_NT.append(False)

                    compostional_rep = self.p_compositional(popped_rep)
                    combinedElement = Element(nt_tree)

                    state.stack_stackrnn.push(compostional_rep, combinedElement)
                elif target_action_idx in self.valid_NT_idxs:

                    # if this is root prediction and if that root is one
                    # of the unsupported intents
                    if (
                        len(state.predicted_actions_idx) == 1
                        and target_action_idx in self.ignore_subNTs_roots
                    ):
                        state.found_unsupported = True

                    state.is_open_NT.append(True)
                    state.num_open_NT += 1
                    state.stack_stackrnn.push(
                        action_embedding, Element(target_action_idx)
                    )
                else:
                    assert "not a valid action: {}".format(
                        self.actions_vocab.itos[target_action_idx]
                    )

                state.neg_prob = neg_prob
                beam.append(state)
            # End for
        # End while
        assert len(beam) > 0, "How come beam is empty?"
        assert len(state.stack_stackrnn) == 1, "How come stack len is " + str(
            len(state.stack_stackrnn)
        )
        assert len(state.buffer_stackrnn) == 0, "How come buffer len is " + str(
            len(state.buffer_stackrnn)
        )

        # Add batch dimension before returning.
        if topk <= 1:
            state = min(beam)
            return (
                torch.LongTensor(state.predicted_actions_idx).unsqueeze(0),
                torch.cat(state.action_scores).unsqueeze(0),
            )
        else:
            return [
                (
                    torch.LongTensor(state.predicted_actions_idx).unsqueeze(0),
                    torch.cat(state.action_scores).unsqueeze(0),
                )
                for state in sorted(beam)[:topk]
            ]

    # Initializes LSTM parameters
    def init_lstm(self):
        return (
            torch.zeros(self.lstm_num_layers, 1, self.lstm_dim),
            torch.zeros(self.lstm_num_layers, 1, self.lstm_dim),
        )

    def valid_actions(self, state: ParserState) -> List[int]:
        """Used for restricting the set of possible action predictions

        Args:
        state : ParserState
            The state of the stack, buffer and action

        Returns:
        List[int]
            indices of the valid actions

        """
        valid_actions: List[int] = []
        is_open_NT = state.is_open_NT
        num_open_NT = state.num_open_NT
        stack = state.stack_stackrnn
        buffer = state.buffer_stackrnn

        # Reduce if there are 1. the top of stack is not an NT, and
        # 2. two open NT on stack, or 3. buffer is empty
        if (is_open_NT and not is_open_NT[-1]) and (
            num_open_NT >= 2 or len(buffer) == 0
        ):
            assert len(stack) > 0
            valid_actions.append(self.reduce_idx)

        if len(buffer) > 0 and num_open_NT < self.max_open_NT:
            last_open_NT = None
            try:
                last_open_NT = stack.ele_from_top(is_open_NT[::-1].index(True))
            except ValueError:
                pass

            if (not self.training) or self.constraints_intent_slot_nesting:
                # if stack is empty or the last open NT is slot
                if (not last_open_NT) or last_open_NT.node in self.valid_SL_idxs:
                    valid_actions += self.valid_IN_idxs
                elif last_open_NT.node in self.valid_IN_idxs:
                    if (
                        self.constraints_no_slots_inside_unsupported
                        and state.found_unsupported
                    ):
                        pass
                    else:
                        valid_actions += self.valid_SL_idxs
            else:
                valid_actions += self.valid_IN_idxs
                valid_actions += self.valid_SL_idxs

        elif (not self.training) and num_open_NT >= self.max_open_NT:
            print(
                "not predicting NT because buffer len is "
                + str(len(buffer))
                + " and num open NTs is "
                + str(num_open_NT)
            )
        if len(buffer) > 0 and num_open_NT >= 1:
            valid_actions.append(self.shift_idx)

        assert (
            len(valid_actions) > 0
        ), "No valid actions. stack is {}, buffer is {}, num open NT: {}".format(
            str(stack), str(buffer), str(num_open_NT)
        )
        return valid_actions

    def get_param_groups_for_optimizer(self):
        """
        This is called by code that looks for an instance of pytext.models.model.Model.
        """
        return [{"params": self.parameters()}]

    def get_loss(
        self, logits: torch.Tensor, target_actions: torch.Tensor, context: torch.Tensor
    ):
        # action scores is a 2D Tensor of dims sequence_length x number_of_actions
        # targets is a 1D list of integers of length sequence_length

        # Get rid of the batch dimension
        action_scores = logits[1].squeeze(0)
        target_actions = target_actions.squeeze(0)

        action_scores_list = torch.chunk(action_scores, action_scores.size()[0])
        target_vars = [
            cuda_utils.Variable(torch.LongTensor([t])) for t in target_actions
        ]
        losses = [
            self.loss_func(action, target).view(1)
            for action, target in zip(action_scores_list, target_vars)
        ]
        total_loss = torch.sum(torch.cat(losses)) if len(losses) > 0 else None
        return total_loss

    def get_pred(self, logits: Tuple[torch.Tensor, torch.Tensor], *args):
        predicted_action_idx, predicted_action_scores = logits
        predicted_scores = [
            np.exp(np.max(action_scores)).item() / np.sum(np.exp(action_scores)).item()
            for action_scores in predicted_action_scores.detach().squeeze(0).tolist()
        ]

        return predicted_action_idx.tolist(), [predicted_scores]

    def save_modules(self, *args, **kwargs):
        pass

    def contextualize(self, context):
        self.context = context



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import os
from typing import Any, Dict, List, Tuple, get_type_hints

import torch
from pytext.config import PyTextConfig, TestConfig
from pytext.config.component import create_exporter
from pytext.metric_reporters.channel import Channel
from pytext.task import Task, create_task, load, save
from pytext.utils.dist_utils import dist_init
from tensorboardX import SummaryWriter

from .utils import cuda_utils


def _set_cuda(
    use_cuda_if_available: bool, device_id: int = 0, world_size: int = 1
) -> None:
    cuda_utils.CUDA_ENABLED = use_cuda_if_available and torch.cuda.is_available()
    cuda_utils.DISTRIBUTED_WORLD_SIZE = world_size

    if use_cuda_if_available and not cuda_utils.CUDA_ENABLED:
        print("Cuda is not available, running on CPU...")
    elif cuda_utils.CUDA_ENABLED:
        torch.cuda.set_device(device_id)

    print(
        """
    # for debug of GPU
    use_cuda_if_available: {}
    device_id: {}
    world_size: {}
    torch.cuda.is_available(): {}
    cuda_utils.CUDA_ENABLED: {}
    cuda_utils.DISTRIBUTED_WORLD_SIZE: {}
    """.format(
            use_cuda_if_available,
            device_id,
            world_size,
            torch.cuda.is_available(),
            cuda_utils.CUDA_ENABLED,
            cuda_utils.DISTRIBUTED_WORLD_SIZE,
        )
    )


def train_model(
    config: PyTextConfig,
    dist_init_url: str = None,
    device_id: int = 0,
    rank: int = 0,
    world_size: int = 1,
) -> Tuple:
    task = prepare_task(config, dist_init_url, device_id, rank, world_size)
    trained_model, best_metric = task.train(config, rank, world_size)
    # Only rank 0 gets to finalize the job and export the model
    if rank == 0:
        save_and_export(config, task)
    return trained_model, best_metric


def prepare_task(
    config: PyTextConfig,
    dist_init_url: str = None,
    device_id: int = 0,
    rank: int = 0,
    world_size: int = 1,
) -> Task:

    if dist_init_url and world_size > 1:
        dist_init(rank, world_size, dist_init_url)

    print("\nParameters: {}\n".format(config))
    _set_cuda(config.use_cuda_if_available, device_id, world_size)
    if config.load_snapshot_path and os.path.isfile(config.load_snapshot_path):
        return load(config.load_snapshot_path)
    return create_task(config.task)


def save_and_export(
    config: PyTextConfig, task: Task, summary_writer: SummaryWriter = None
) -> None:
    print("\n=== Saving model to: " + config.save_snapshot_path)
    save(config, task.model, task.data_handler.metadata_to_save())
    task.export(task.model, config.export_caffe2_path, summary_writer)


def export_saved_model_to_caffe2(
    saved_model_path: str, export_caffe2_path: str
) -> None:
    task, train_config = load(saved_model_path)
    if task.exporter is None:
        TaskType = type(train_config.task)
        ExporterConfigType = get_type_hints(TaskType)["exporter"].__args__[0]
        task.exporter = create_exporter(
            ExporterConfigType(),
            train_config.task.features,
            train_config.task.labels,
            task.data_handler.metadata,
        )
    task.export(task.model, export_caffe2_path)


def test_model(test_config: TestConfig, metrics_channel: Channel = None) -> Any:
    _set_cuda(test_config.use_cuda_if_available)

    task, train_config = load(test_config.load_snapshot_path)
    if metrics_channel is not None:
        task.metric_reporter.add_channel(metrics_channel)

    return (
        task.test(test_config.test_path),
        train_config.task.metric_reporter.output_path,
    )


def batch_predict(model_file: str, examples: List[Dict[str, Any]]):
    task, train_config = load(model_file)
    return task.predict(examples)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from .loss import AUCPRHingeLoss, BinaryCrossEntropyLoss, CrossEntropyLoss, Loss


__all__ = ["Loss", "CrossEntropyLoss", "BinaryCrossEntropyLoss", "AUCPRHingeLoss"]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import torch
from pytext.config.field_config import PretrainedModelEmbeddingConfig

from .embedding_base import EmbeddingBase


class PretrainedModelEmbedding(EmbeddingBase):
    """Module for providing token embeddings from a pretrained model."""

    Config = PretrainedModelEmbeddingConfig

    @classmethod
    def from_config(cls, config: PretrainedModelEmbeddingConfig, *args, **kwargs):
        return cls(config.embed_dim)

    def forward(self, embedding: torch.Tensor) -> torch.Tensor:
        if embedding.shape[2] != self.embedding_dim:
            raise ValueError(
                f"Expected {self.embedding_dim} as dimension for pretrained "
                + f"model embedding but got {embedding.shape[2]}."
            )
        return embedding



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from pytext.common.constants import BatchContext

from .metric_reporter import MetricReporter


class DisjointMultitaskMetricReporter(MetricReporter):
    def __init__(self, reporters) -> None:
        super().__init__(None)
        self.reporters = reporters

    def add_batch_stats(
        self, n_batches, preds, targets, scores, loss, m_input, **context
    ):
        reporter = self.reporters[context[BatchContext.TASK_NAME]]
        reporter.add_batch_stats(
            n_batches, preds, targets, scores, loss, m_input, **context
        )

    def add_channel(self, channel):
        for reporter in self.reporters.values():
            reporter.add_channel(channel)

    def report_metric(self, stage, epoch, reset=True, print_to_channels=True):
        metrics_dict = {}
        for name, reporter in self.reporters.items():
            print(f"Reporting on task: {name}")
            metrics_dict[name] = reporter.report_metric(
                stage, epoch, reset, print_to_channels
            )
        return metrics_dict

    def compare_metric(self, new_metric, old_metric):
        return True



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import tempfile
import unittest

import numpy as np
from pytext.common.constants import DatasetFieldName
from pytext.config.field_config import (
    DocLabelConfig,
    EmbedInitStrategy,
    FeatureConfig,
    WordFeatConfig,
    WordLabelConfig,
)
from pytext.data import JointModelDataHandler
from pytext.data.featurizer import SimpleFeaturizer
from pytext.utils.embeddings_utils import PretrainedEmbedding
from pytext.utils.test_utils import import_tests_module


tests_module = import_tests_module()
TRAIN_FILE = tests_module.test_file("train_data_tiny.tsv")
EVAL_FILE = tests_module.test_file("test_data_tiny.tsv")
TEST_FILE = tests_module.test_file("test_data_tiny.tsv")


EMBED_RAW_PATH = tests_module.test_file("pretrained_embed_raw")
EMBED_CACHED_PATH = tests_module.test_file("test_embed.cached")
EMBED_XLU_CACHED_PATH = tests_module.test_file("test_embed_xlu.cached")


class PretrainedEmbedsTest(unittest.TestCase):
    def test_cache_embeds(self):
        embeddings_ref = PretrainedEmbedding()
        embeddings_ref.load_pretrained_embeddings(EMBED_RAW_PATH)
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=".{}".format("cached")
        ) as cached_path:
            embeddings_ref.cache_pretrained_embeddings(cached_path.name)
            embeddings_cached = PretrainedEmbedding()
            embeddings_cached.load_cached_embeddings(cached_path.name)

        np.testing.assert_array_equal(
            sorted(embeddings_cached.stoi.keys()), sorted(embeddings_ref.stoi.keys())
        )
        np.testing.assert_array_equal(
            embeddings_cached.embed_vocab, embeddings_ref.embed_vocab
        )
        np.testing.assert_array_equal(
            sorted(embeddings_cached.stoi.values()),
            sorted(embeddings_ref.stoi.values()),
        )
        for word_idx in embeddings_ref.stoi.values():
            np.testing.assert_array_almost_equal(
                embeddings_cached.embedding_vectors[word_idx],
                embeddings_ref.embedding_vectors[word_idx],
            )

    def test_assing_pretrained_weights(self):
        embeddings_ref = PretrainedEmbedding()
        embeddings_ref.load_cached_embeddings(EMBED_CACHED_PATH)
        VOCAB = ["UNK", "aloha", "the"]
        EMBED_DIM = 5
        # Get Vocab to Idx:
        UNK_IDX = 0
        embed_vocab_to_idx = {}
        for word in embeddings_ref.embed_vocab:
            if word in VOCAB:
                embed_vocab_to_idx[word] = VOCAB.index(word)
            else:
                embed_vocab_to_idx[word] = UNK_IDX
        pretrained_embeds = embeddings_ref.initialize_embeddings_weights(
            embed_vocab_to_idx, UNK_IDX, len(VOCAB), EMBED_DIM, EmbedInitStrategy.RANDOM
        )
        assert pretrained_embeds.shape[0] == len(VOCAB)
        assert pretrained_embeds.shape[1] == EMBED_DIM
        np.testing.assert_array_almost_equal(
            pretrained_embeds[1].numpy(),
            [-0.43124, 0.014934, -0.50635, 0.60506, 0.56051],
        )  # embedding vector for 'aloha'
        np.testing.assert_array_almost_equal(
            pretrained_embeds[2].numpy(),
            [-0.39153, -0.19803, 0.2573, -0.18617, 0.25551],
        )  # embedding vector for 'the'

    def test_cache_xlu_embeds(self):
        embeddings_ref = PretrainedEmbedding()

        dialects = ["en_US", "en_UK", "es_XX"]
        for dialect in dialects:
            embeddings_ref.load_pretrained_embeddings(
                EMBED_RAW_PATH, append=True, dialect=dialect
            )
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=".{}".format("cached")
        ) as cached_path:
            embeddings_ref.cache_pretrained_embeddings(cached_path.name)
            embeddings_cached = PretrainedEmbedding()
            embeddings_cached.load_cached_embeddings(cached_path.name)

        np.testing.assert_array_equal(
            sorted(embeddings_cached.stoi.keys()), sorted(embeddings_ref.stoi.keys())
        )
        np.testing.assert_array_equal(
            embeddings_cached.embed_vocab, embeddings_ref.embed_vocab
        )
        np.testing.assert_array_equal(
            sorted(embeddings_cached.stoi.values()),
            sorted(embeddings_ref.stoi.values()),
        )
        for word_idx in embeddings_ref.stoi.values():
            np.testing.assert_array_almost_equal(
                embeddings_cached.embedding_vectors[word_idx],
                embeddings_ref.embedding_vectors[word_idx],
            )

    def test_assing_pretrained_xlu_weights(self):
        embeddings_ref = PretrainedEmbedding()
        embeddings_ref.load_cached_embeddings(EMBED_XLU_CACHED_PATH)
        VOCAB = ["UNK", "aloha-en_US", "the-es_XX"]
        EMBED_DIM = 5
        # Get Vocab to Idx:
        UNK_IDX = 0
        embed_vocab_to_idx = {}
        for word in embeddings_ref.embed_vocab:
            if word in VOCAB:
                embed_vocab_to_idx[word] = VOCAB.index(word)
            else:
                embed_vocab_to_idx[word] = UNK_IDX
        pretrained_embeds = embeddings_ref.initialize_embeddings_weights(
            embed_vocab_to_idx, UNK_IDX, len(VOCAB), EMBED_DIM, EmbedInitStrategy.RANDOM
        )
        assert pretrained_embeds.shape[0] == len(VOCAB)
        assert pretrained_embeds.shape[1] == EMBED_DIM
        np.testing.assert_array_almost_equal(
            pretrained_embeds[1].numpy(),
            [-0.43124, 0.014934, -0.50635, 0.60506, 0.56051],
        )  # embedding vector for 'aloha-en_US'
        np.testing.assert_array_almost_equal(
            pretrained_embeds[2].numpy(),
            [-0.39153, -0.19803, 0.2573, -0.18617, 0.25551],
        )  # embedding vector for 'the-es_XX'

    def test_intializing_embeds_from_config(self):
        feature_config = FeatureConfig(
            word_feat=WordFeatConfig(
                embedding_init_strategy=EmbedInitStrategy.RANDOM,
                embed_dim=5,
                pretrained_embeddings_path=tests_module.TEST_BASE_DIR,
            )
        )
        data_handler = JointModelDataHandler.from_config(
            JointModelDataHandler.Config(),
            feature_config,
            [DocLabelConfig(), WordLabelConfig()],
            featurizer=SimpleFeaturizer.from_config(
                SimpleFeaturizer.Config(), feature_config
            ),
        )

        data_handler.init_metadata_from_path(TRAIN_FILE, EVAL_FILE, TEST_FILE)

        pretrained_embeds = data_handler.metadata.features[
            DatasetFieldName.TEXT_FIELD
        ].pretrained_embeds_weight
        # test random initialization (values should be non-0)
        np.testing.assert_array_less(
            [0, 0, 0, 0, 0], np.absolute(pretrained_embeds[11].numpy())
        )

        feature_config = FeatureConfig(
            word_feat=WordFeatConfig(
                embedding_init_strategy=EmbedInitStrategy.ZERO,
                embed_dim=5,
                pretrained_embeddings_path=tests_module.TEST_BASE_DIR,
            )
        )
        data_handler = JointModelDataHandler.from_config(
            JointModelDataHandler.Config(),
            feature_config,
            [DocLabelConfig(), WordLabelConfig()],
            featurizer=SimpleFeaturizer.from_config(
                SimpleFeaturizer.Config(), feature_config
            ),
        )
        data_handler.init_metadata_from_path(TRAIN_FILE, EVAL_FILE, TEST_FILE)

        pretrained_embeds = data_handler.metadata.features[
            DatasetFieldName.TEXT_FIELD
        ].pretrained_embeds_weight
        # test zero initialization (values should all be 0)
        np.testing.assert_array_equal([0, 0, 0, 0, 0], pretrained_embeds[11].numpy())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from typing import List

import torch
from pytext.config.pytext_config import OptimizerParams, OptimizerType
from pytext.models.model import Model


def create_optimizer(
    model: Model, optimizer_params: OptimizerParams
) -> List[torch.optim.Optimizer]:
    if optimizer_params.type == OptimizerType.ADAM:
        return [
            torch.optim.Adam(
                model.get_param_groups_for_optimizer(),
                lr=optimizer_params.lr,
                weight_decay=optimizer_params.weight_decay,
            )
        ]
    elif optimizer_params.type == OptimizerType.SGD:
        return [
            torch.optim.SGD(
                model.get_param_groups_for_optimizer(),
                lr=optimizer_params.lr,
                momentum=optimizer_params.momentum,
            )
        ]
    else:
        raise ValueError("Unknown optimizer type")


def optimizer_zero_grad(optimizers: List[torch.optim.Optimizer]) -> None:
    for op in optimizers:
        op.zero_grad()


def optimizer_step(optimizers: List[torch.optim.Optimizer]) -> None:
    for op in optimizers:
        op.step()


def learning_rates(optimizers):
    for optimizer in optimizers:
        for param_group in optimizer.param_groups:
            yield param_group["lr"]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from multiprocessing import Manager
from typing import List

import torch
import torch.multiprocessing as mp
from pytext.config import PyTextConfig
from pytext.config.component import create_trainer
from pytext.config.pytext_config import ConfigBase
from pytext.metric_reporters import MetricReporter
from pytext.models.model import Model
from pytext.trainers.trainer import Trainer
from torchtext.data import Iterator


class HogwildTrainer(Trainer):
    class Config(ConfigBase):
        real_trainer: Trainer.Config = Trainer.Config()
        num_workers: int = 1

    @classmethod
    def from_config(cls, config: Config, *args, **kwargs):
        return cls(
            create_trainer(config.real_trainer, *args, **kwargs), config.num_workers
        )

    def __init__(self, real_trainer, num_workers):
        self.real_trainer = real_trainer
        self.num_workers = num_workers

    def test(self, model, test_iter, metric_reporter):
        return self.real_trainer.test(model, test_iter, metric_reporter)

    def train(
        self,
        train_iter: Iterator,
        eval_iter: Iterator,
        model: Model,
        metric_reporter: MetricReporter,
        optimizers: List[torch.optim.Optimizer],
        pytext_config: PyTextConfig,
        scheduler=None,
        *args,
        **kwargs
    ):
        print("Num of workers for Hogwild Training is {}".format(self.num_workers))

        # Share memory of tensors for concurrent updates from multiple processes.
        if self.num_workers > 1:
            for param in model.parameters():
                param.share_memory_()

        processes = []
        for rank in range(1, self.num_workers):
            # Initialize the batches with different randome states.
            train_iter.batches.init_epoch()
            p = mp.Process(
                target=self.real_trainer.train,
                args=(
                    train_iter,
                    eval_iter,
                    model,
                    metric_reporter,
                    optimizers,
                    pytext_config,
                    scheduler,
                    None,
                    rank,
                ),
            )
            processes.append(p)
            p.start()

        training_result: List = Manager().list()  # Actual type is ListProxy.
        self.real_trainer.train(
            train_iter,
            eval_iter,
            model,
            metric_reporter,
            optimizers,
            pytext_config,
            scheduler,
            training_result,
            rank=0,
        )

        for p in processes:
            p.join()

        # Ony rank 0 worker writes to training_result
        assert len(training_result) == 1
        return training_result[0]  # Contains best model and best metric.



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import itertools
from collections import OrderedDict
from typing import Dict, Optional, Tuple

from pytext.common.constants import BatchContext

from .data_handler import BatchIterator, DataHandler


class RoundRobinBatchIterator(BatchIterator):
    """We take a dictionary of BatchIterators and do round robin over
    them in a cycle.  If epoch_size is specified, each iterator is also
    wrapped in a cycle so that they never run out.  Otherwise, a single
    pass is done over each iterator at each epoch.  Iterators that run
    out are filtered out.  Currently there is no re-shuffling of data,
    data order is the same at each epoch.

    e.g.  Iterator 1: [A, B, C, D],  Iterator 2: [a, b]
    Case 1, epoch size is set:
        Output: [A, a, B, b, C, a, D, b, A, ...]
        Here, tasks with less data are effectively upsampled and data is
        balanced accross tasks.
    Case 2, epoch size not set:
        Output: [A, a, B, b, C, D, A, a, B, b, ...]

    Args:
        iterators (Dict[str, BatchIterator]): Iterators to do roundrobin over.
        epoch_size (Optional[int]): Size of epoch in number of batches.  If not set,
        do a single pass over the training data.

    Attributes:
        iterators (type): Iterators to do roundrobin over.
        epoch_size (type): Size of epoch in number of batches.

    """

    def __init__(
        self, iterators: Dict[str, BatchIterator], epoch_size: Optional[int] = None
    ) -> None:
        self.iterators = iterators
        self.epoch_size = epoch_size or float("inf")

    def __iter__(self):
        iterators = {
            name: iter(
                self.cycle(iterator) if (self.epoch_size < float("inf")) else iterator
            )
            for name, iterator in self.iterators.items()
        }

        round_robin = itertools.filterfalse(  # filter iterators that run out
            lambda x: not bool(x),
            # chain list of tuples, resulting in round robin
            itertools.chain.from_iterable(
                # zip list of iterators,
                # return tuples with one element from each iterator
                itertools.zip_longest(
                    *[  # turn into iterator of (name, batch) tuples
                        zip(itertools.repeat(name), iterator)
                        for name, iterator in iterators.items()
                    ]
                )
            ),
        )

        for i, (name, (input, target, context)) in enumerate(round_robin):
            if i >= self.epoch_size:
                # end of epoch
                return
            context[BatchContext.TASK_NAME] = name
            yield input, target, context

    @classmethod
    def cycle(cls, iterator):
        while True:
            for item in iterator:
                yield item


class DisjointMultitaskDataHandler(DataHandler):
    """
    Wrapper for doing multitask training using multiple data handlers.
    Takes a dictionary of data handlers, does round robin over their
    iterators using RoundRobinBatchIterator.

    Args:
        config (Config): Configuration object of type DisjointMultitaskDataHandler.Config.
        data_handlers (Dict[str, DataHandler]): Data handlers to do roundrobin over.
        *args (type): Extra arguments to be passed down to sub data handlers.
        **kwargs (type): Extra arguments to be passed down to sub data handlers.

    Attributes:
        data_handlers (type): Data handlers to do roundrobin over.
        epoch_size (type): Size of epoch in number of batches.
        epoch_size (Optional[int]): Size of epoch in number of batches.  If not set,
        do a single pass over the training data.

    """

    class Config(DataHandler.Config):
        """Configuaration class for `DisjointMultitaskDataHandler`.

        Attributes:
            epoch_size (Optional[int]): Size of epoch in number of batches.  If not set,
            do a single pass over the training data.

        """

        epoch_size: Optional[int] = 10

    def __init__(
        self, config: Config, data_handlers: Dict[str, DataHandler], *args, **kwargs
    ) -> None:
        super(DisjointMultitaskDataHandler, self).__init__(config, None, None, None)
        self.data_handlers = data_handlers
        self.epoch_size = config.epoch_size if config.epoch_size else None

    def get_train_iter(
        self, rank: int = 0, world_size: int = 1
    ) -> Tuple[BatchIterator, ...]:
        iterators: Dict = OrderedDict(
            (name, data_handler.get_train_iter(rank, world_size))
            for name, data_handler in self.data_handlers.items()
        )
        return RoundRobinBatchIterator(iterators, epoch_size=self.epoch_size)

    def get_eval_iter(self) -> BatchIterator:
        iterators: Dict = OrderedDict(
            (name, data_handler.get_eval_iter())
            for name, data_handler in self.data_handlers.items()
        )
        return RoundRobinBatchIterator(iterators)

    def get_test_iter(self) -> BatchIterator:
        iterators: Dict = OrderedDict(
            (name, data_handler.get_test_iter())
            for name, data_handler in self.data_handlers.items()
        )
        return RoundRobinBatchIterator(iterators)

    def init_metadata(self):
        # get data sets
        self.metadata = {}
        for name, data_handler in self.data_handlers.items():
            data_handler.init_metadata()
            self.metadata[name] = data_handler.metadata

    def load_metadata(self, metadata):
        self.metadata = metadata
        for name, data_handler in self.data_handlers.items():
            data_handler.load_metadata(metadata[name])

    def metadata_to_save(self):
        metadata = {}
        for name, data_handler in self.data_handlers.items():
            metadata[name] = data_handler.metadata_to_save()
        return metadata



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from collections import OrderedDict
from pprint import pprint
from typing import Dict

from pytext.config import config_to_json
from pytext.config.component import (
    create_data_handler,
    create_featurizer,
    create_metric_reporter,
    create_model,
    create_trainer,
)
from pytext.data import DisjointMultitaskDataHandler
from pytext.metric_reporters.disjoint_multitask_metric_reporter import (
    DisjointMultitaskMetricReporter,
)
from pytext.models.disjoint_multitask_model import DisjointMultitaskModel
from pytext.optimizer import create_optimizer
from pytext.optimizer.scheduler import Scheduler
from pytext.utils import cuda_utils

from . import Task, TaskBase


class DisjointMultitask(TaskBase):
    """Modules which have the same shared_module_key and type share parameters.
       Only the first instance of such module should be configured in tasks list.
    """

    class Config(TaskBase.Config):
        tasks: Dict[str, Task.Config]
        data_handler: DisjointMultitaskDataHandler.Config = DisjointMultitaskDataHandler.Config()
        metric_reporter: DisjointMultitaskMetricReporter.Config = DisjointMultitaskMetricReporter.Config()

    @classmethod
    def from_config(cls, task_config, metadata=None, model_state=None):
        print("Task parameters:\n")
        pprint(config_to_json(type(task_config), task_config))

        data_handlers = OrderedDict()
        for name, task in task_config.tasks.items():
            featurizer = create_featurizer(task.featurizer, task.features)
            data_handlers[name] = create_data_handler(
                task.data_handler, task.features, task.labels, featurizer=featurizer
            )
        data_handler = DisjointMultitaskDataHandler(
            task_config.data_handler, data_handlers
        )
        print("\nLoading data...")
        if metadata:
            data_handler.load_metadata(metadata)
        else:
            data_handler.init_metadata()
        metadata = data_handler.metadata

        metric_reporter = DisjointMultitaskMetricReporter(
            OrderedDict(
                (name, create_metric_reporter(task.metric_reporter, metadata[name]))
                for name, task in task_config.tasks.items()
            )
        )

        model = DisjointMultitaskModel(
            OrderedDict(
                (name, create_model(task.model, task.features, metadata[name]))
                for name, task in task_config.tasks.items()
            )
        )
        if model_state:
            model.load_state_dict(model_state)
        if cuda_utils.CUDA_ENABLED:
            model = model.cuda()

        optimizers = create_optimizer(model, task_config.optimizer)
        return cls(
            trainer=create_trainer(task_config.trainer),
            data_handler=data_handler,
            model=model,
            metric_reporter=metric_reporter,
            optimizers=optimizers,
            lr_scheduler=Scheduler(
                optimizers, task_config.scheduler, metric_reporter.lower_is_better
            ),
            exporter=None,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from typing import Any, List, Sized, Tuple

import torch as torch
import torch.nn as nn
from pytext.utils.cuda_utils import xaviervar


# token/non-terminal/sub-tree element on a stack.
# need this value for computing valid actions
class Element:
    def __init__(self, node) -> None:
        self.node = node

    def __str__(self):
        return str(self.node)

    def __repr__(self):
        return self.__str__()


class StackLSTM(Sized):
    def __init__(self, rnn, initial_state, p_empty_embedding):
        self.rnn = rnn
        self.list = (
            [(initial_state, (self._rnn_get_output(initial_state), "Root"))]
            if initial_state
            else None
        )
        self.empty = p_empty_embedding

    def _rnn_get_output(self, state):
        return state[0][-1]

    def push(self, expr, ele: Element) -> None:
        # assuming expr is always one element at a time; not a sequence.
        # making it a sequence since LSTM takes a sequence
        expr = expr.unsqueeze(1)
        output, new_embedding = self.rnn(expr, self.list[-1][0])
        self.list.append((new_embedding, (self._rnn_get_output(new_embedding), ele)))

    def pop(self) -> Tuple[Any, Element]:
        # returning tuple of out embedding and the name of the element
        return self.list.pop()[1]

    def top(self) -> Tuple[Any, Element]:
        return self.list[-1][1]

    def embedding(self):
        return (
            self._rnn_get_output(self.list[-1][0]) if len(self.list) > 1 else self.empty
        )

    def first_ele_match(self, funct):
        for st in self.list[::-1]:
            if funct(st):
                return st[1][1]
        return None

    def ele_from_top(self, index: int) -> Element:
        return self.list[len(self.list) - index - 1][1][1]

    def __len__(self):
        return len(self.list) - 1

    def __str__(self):
        return "->".join([str(x[1][1]) for x in self.list])

    def copy(self):
        other = StackLSTM(self.rnn, None, self.empty)
        other.list = list(self.list)
        return other


class CompositionFunction(nn.Module):
    def __init__(self):
        super().__init__()


class CompositionalNN(CompositionFunction):
    def __init__(self, lstm_dim):
        super().__init__()
        self.lstm_dim = lstm_dim
        self.lstm_fwd = nn.LSTM(lstm_dim, lstm_dim, 1)
        self.lstm_rev = nn.LSTM(lstm_dim, lstm_dim, 1)
        self.linear_seq = nn.Sequential(nn.Linear(2 * lstm_dim, lstm_dim), nn.Tanh())

    def forward(self, x):
        """
        Embed the sequence. If the input corresponds to [IN:GL where am I at]:
        - x will contain the embeddings of [at I am where IN:GL] in that order.
        - Forward LSTM will embed the sequence [IN:GL where am I at].
        - Backward LSTM will embed the sequence [IN:GL at I am where].
        The final hidden states are concatenated and then projected.

        Args:
            x: Embeddings of the input tokens in *reversed* order
        """
        # reset hidden every time
        lstm_hidden_fwd = (
            xaviervar(1, 1, self.lstm_dim),
            xaviervar(1, 1, self.lstm_dim),
        )
        lstm_hidden_rev = (
            xaviervar(1, 1, self.lstm_dim),
            xaviervar(1, 1, self.lstm_dim),
        )
        nt_element = x[-1]
        rev_rest = x[:-1]
        # Always put nt_element at the front
        fwd_input = [nt_element] + rev_rest[::-1]
        rev_input = [nt_element] + rev_rest
        stacked_fwd = self.lstm_fwd(torch.stack(fwd_input), lstm_hidden_fwd)[0][0]
        stacked_rev = self.lstm_rev(torch.stack(rev_input), lstm_hidden_rev)[0][0]
        combined = torch.cat([stacked_fwd, stacked_rev], dim=1)
        subtree_embedding = self.linear_seq(combined)
        return subtree_embedding


class CompositionalSummationNN(CompositionFunction):
    def __init__(self, lstm_dim):
        super().__init__()
        self.lstm_dim = lstm_dim

        self.linear_seq = nn.Sequential(nn.Linear(lstm_dim, lstm_dim), nn.Tanh())

    def forward(self, x):
        combined = torch.sum(torch.cat(x, dim=0), dim=0, keepdim=True)
        subtree_embedding = self.linear_seq(combined)
        return subtree_embedding


class ParserState:
    # Copies another state instead if supplied
    def __init__(self, parser=None):

        if not parser:
            return
        # Otherwise initialize normally
        self.buffer_stackrnn = StackLSTM(
            parser.buff_rnn, parser.init_lstm(), parser.pempty_buffer_emb
        )
        self.stack_stackrnn = StackLSTM(
            parser.stack_rnn, parser.init_lstm(), parser.empty_stack_emb
        )
        self.action_stackrnn = StackLSTM(
            parser.action_rnn, parser.init_lstm(), parser.empty_action_emb
        )

        self.predicted_actions_idx = []
        self.action_scores = []

        self.num_open_NT = 0
        self.is_open_NT: List[bool] = []
        self.found_unsupported = False

        # negative cumulative log prob so sort(states) is in descending order
        self.neg_prob = 0

    def finished(self):
        return len(self.stack_stackrnn) == 1 and len(self.buffer_stackrnn) == 0

    def copy(self):
        other = ParserState()
        other.buffer_stackrnn = self.buffer_stackrnn.copy()
        other.stack_stackrnn = self.stack_stackrnn.copy()
        other.action_stackrnn = self.action_stackrnn.copy()
        other.predicted_actions_idx = self.predicted_actions_idx.copy()
        other.action_scores = self.action_scores.copy()
        other.num_open_NT = self.num_open_NT
        other.is_open_NT = self.is_open_NT.copy()
        other.neg_prob = self.neg_prob
        other.found_unsupported = self.found_unsupported
        return other

    def __gt__(self, other):
        return self.neg_prob > other.neg_prob

    def __eq__(self, other):
        return self.neg_prob == other.neg_prob



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import re
from typing import List, Optional, Sequence, Tuple

from pytext.common.constants import VocabMeta
from pytext.config import ConfigBase
from pytext.data.featurizer import Featurizer, InputRecord, OutputRecord


class SimpleFeaturizer(Featurizer):
    """
    Simple featurizer for basic tokenization and gazetteer feature alignment.
    """

    class Config(ConfigBase):
        sentence_markers: Optional[Tuple[str, str]] = None
        lowercase_tokens: bool = True
        split_regex: str = r"\s+"

    def tokenize(self, input_record: InputRecord) -> OutputRecord:
        """Tokenize one instance/example only."""
        tokens: List[str] = []
        token_ranges: List[Tuple[int]] = []

        def add_token(text, start, end):
            token = text[start:end]
            if token:
                tokens.append(token)
                token_ranges.append((start, end))

        start = 0
        text = input_record.raw_text
        for sep in re.finditer(self.config.split_regex, text):
            add_token(text, start, sep.start())
            start = sep.end()
        add_token(text, start, len(text))

        if not tokens:
            # Add PAD_TOKEN in case of empty text
            tokens = [VocabMeta.PAD_TOKEN]
        if self.config.lowercase_tokens:
            tokens = list(map(str.lower, tokens))
        if self.config.sentence_markers:
            tokens.insert(0, self.config.sentence_markers[0])
            tokens.append(self.config.sentence_markers[1])

        return OutputRecord(tokens=tokens, token_ranges=token_ranges)

    def tokenize_batch(
        self, input_records: Sequence[InputRecord]
    ) -> Sequence[OutputRecord]:
        return [self.tokenize(in_record) for in_record in input_records]

    def featurize(self, input_record: InputRecord) -> OutputRecord:
        """Featurize one instance/example only."""
        return self.tokenize(input_record)

    def featurize_batch(
        self, input_records: Sequence[InputRecord]
    ) -> Sequence[OutputRecord]:
        return [self.featurize(in_record) for in_record in input_records]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from typing import List

from pytext.common.constants import Stage
from pytext.data import CommonMetadata
from pytext.metrics import LabelPrediction, compute_classification_metrics

from .channel import Channel, ConsoleChannel, FileChannel
from .metric_reporter import MetricReporter


class IntentModelChannel(FileChannel):
    def get_title(self):
        return ("predicted", "actual", "scores_str", "text")

    def gen_content(self, metrics, loss, preds, targets, scores, contexts):
        for i in range(len(preds)):
            yield [
                preds[i],
                targets[i],
                ",".join([f"{s:.2f}" for s in scores[i]]),
                contexts["utterance"][i],
            ]


class ClassificationMetricReporter(MetricReporter):
    def __init__(self, label_names: List[str], channels: List[Channel]) -> None:
        super().__init__(channels)
        self.label_names = label_names

    @classmethod
    def from_config(cls, config, meta: CommonMetadata):
        label_names = meta.target.vocab.itos
        return cls(
            label_names,
            [ConsoleChannel(), IntentModelChannel((Stage.TEST,), config.output_path)],
        )

    def calculate_metric(self):
        return compute_classification_metrics(
            [
                LabelPrediction(scores, pred, expect)
                for scores, pred, expect in zip(
                    self.all_scores, self.all_preds, self.all_targets
                )
            ],
            self.label_names,
        )

    def get_meta(self):
        return {"label_names": self.label_names}

    @staticmethod
    def get_model_select_metric(metrics):
        return metrics.accuracy
